window.Gleam.initSearch([{"doc":"glam","title":"glam","content":"![Glam](https://github.com/giacomocavalieri/glam/blob/main/resources/glam_banner.png?raw=true)\n\n[![Package Version](https://img.shields.io/hexpm/v/glam?color=92DCE5)](https://hex.pm/packages/glam)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3?color=FCC0D2)](https://hexdocs.pm/glam/)\n\n✨ A package to help you easily pretty print structured data\n\n&gt; ⚙️ This package works with both the Erlang and Javascript target\n\n## Installation\n\nTo add this package to your Gleam project:\n\n```sh\ngleam add glam\n```\n\n## Pretty printing\n\nWhen working with structured data (like JSON, XML, lists, ASTs, ...) it can\nbe useful to print it in a nice and tidy way.\nOf course, one could go the quick and dirty way and just `string.inspect`\nthe data structure and call it a day.\nHowever, the result would hardly be readable for complex data structures.\n\nOn the other hand, handwriting a pretty printing function can be quite\ndifficult (I sure have spent my fair share of time writing those) and\nerror-prone: there are a lot of moving pieces to juggle and it&#39;s easy for bugs\nto sneak in.\n\nA pretty printing package like Glam provides some basic building blocks to\n_describe_ the structure of your data and takes care of the heavy lifting of\nactually finding the best layout to format the data, gracefully handling line\nlimits and indentation.\n\n## Getting started\n\nIf you want to get started with Glam, a great starting point is the\n[learning materials page](https://hexdocs.pm/glam/learning_materials.html)!\nThere you&#39;ll find a lot of step-by-step tutorials that will guide you through\nsome awesome pretty printers and get you up and running with Glam.\n\nBy the end of the tutorials you&#39;ll know most of the package&#39;s API and you&#39;ll\nhave implemented pretty printers for lists, JSON and error messages:\n\n```gleam\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;])\n|&gt; doc.to_string(10) // Fit the list on lines of width 10\n\n// [\n//   &quot;Gleam&quot;,\n//   &quot;is&quot;,   \n//   &quot;fun!&quot;,   \n// ]\n```\n\nYou can also have a look at the\n[`examples`](https://github.com/giacomocavalieri/glam/tree/main/src/examples)\nfolder in the GitHub repo, there you&#39;ll also find a JSON pretty printer.\n\n## Contributing\n\nIf you think there&#39;s any way to improve this package, or if you spot a bug\ndon&#39;t be afraid to open PRs, issues or requests of any kind!\nAny contribution is welcome 💜\n","url":"index.html"},{"doc":"glam","title":"glam","content":"# Learning materials\n\nPretty printing with a package like Glam can be quite confusing at first: it may\ntake some time to get used to the API, and when writing your first pretty\nprinters it may be difficult to get the desired look.\n\nDon&#39;t be scared, that&#39;s normal! The topic is quite challenging, but that&#39;s\nnothing that some practice cannot fix. There are plenty of learning materials to\nget you started, and all the tutorials will walk you through many examples.\n\nIf you still find the package has some rough edges that could be improved, if\nyou feel like you&#39;re still not getting how something works, if you&#39;re having any\nkind of problem with Glam, or even if you just feel like having a chat... please\n_do reach out_! I&#39;d love to help and your feedback is always appreciated 💜\n\nYou can find me on [Twitter](https://twitter.com/giacomo_cava), hanging around\nthe [Gleam Discord server](https://discord.gg/Fm8Pwmy), and I always check the\nissues and discussions in the\n[Glam repo](https://github.com/giacomocavalieri/glam).\n\n## Step-by-step tutorials\n\nThe best way to get started is by reading the Glam tutorials that will guide you\nthrough some example projects to showcase the library API.\n\nYou will learn how pretty printing works and write some glamorous pretty\nprinters for a variety of data structures: simple string lists, todo lists and\nJSON.\n\nI suggest you work through the tutorials in this order:\n\n- [Pretty printing glamorous lists](https://hexdocs.pm/glam/pretty_printing_glamorous_lists.html):\n  this introductory tutorial will get you up and running with Glam and showcase\n  most of the package API. There&#39;s also a\n  [live stream I did](https://www.youtube.com/watch?v=jxrxr9lH088)\n- [Pretty printing TODO lists](https://hexdocs.pm/glam/pretty_printing_todo_lists.html):\n  this tutorial goes deeper into how nesting works, and introduces another\n  neat function that was omitted in the first tutorial\n- [Pretty printing JSON](https://hexdocs.pm/glam/pretty_printing_JSON.html):\n  this tutorial walks through the implementation of a JSON pretty printer. It\n  won&#39;t introduce any new concepts, but it can be a nice exercise to hone your\n  pretty printing skills\n- [Pretty printing lovely error messages](https://hexdocs.pm/glam/pretty_printing_lovely_error_messages.html):\n  this tutorial will walk you through the implementation of a pretty printer for\n  the error messages of a made-up programming language. You&#39;ll learn about\n  another function that was omitted in the introductory tutorial that can help\n  you easily build beautiful pretty printers\n\n## Full code examples\n\nIf step-by-step tutorials are not your thing, you can still have a look at the\nfull code examples in the\n[project repository](https://github.com/giacomocavalieri/glam/tree/main/src/examples).\n\n## Package API\n\nOf course you can also have a look at the\n[package API](https://hexdocs.pm/glam/glam/doc.html): every function is\ncommented and has a small example to showcase how it can be used.\n","url":"learning_materials.html"},{"doc":"glam","title":"glam","content":"# Pretty printing glamorous lists\n\n## Goals\n\nThe goal of this tutorial is to get you started on pretty printing with the\nGlam package introducing its API a step at a time.\n\nAs a running example, we&#39;re going to write a pretty printer for lists of strings.\nThere are three rules we want our pretty printer to follow:\n\n- If a list is shorter than the maximum allowed line width, it is kept on\n  a single line\n- If it is longer than the maximum line width, each element is placed on a new\n  line and is indented by two spaces; in addition, the last element has a\n  trailing comma\n- All strings are surrounded by double quotes `&quot;`\n\nSince a ~~picture~~ code snippet is worth a thousand words let me show you an\nexample:\n\n```gleam\n// ----------------- // &lt;- This is to show the 20 chars limit\n\n[&quot;Hello&quot;, &quot;Gleam&quot;]   // &lt;- This can stay on a single line, easy!\n\n[                    // &lt;- This wouldn&#39;t fit on a single line\n  &quot;Gleam&quot;,           //    so each element is split on a different   \n  &quot;is&quot;,              //    line and is indented by two spaces \n  &quot;fun!&quot;,            // &lt;- Notice the trailing comma!\n]\n```\n\nNow that we have a clear picture of what a pretty printed list should look\nlike we can dive deep into the wonders of pretty printing!\n\n## Setup\n\nThis tutorial is full of code examples so you can follow along with your own\nGleam project. If you have a working installation of Gleam, open your terminal\nand enter the following:\n\n```shell\ngleam new glam_tutorial # create a new project named `glam_tutorial`\ncd glam_tutorial        # jump into its directory\ngleam add glam          # add the `glam` package to the project&#39;s dependencies\n```\n\n&gt; If you know nothing about Gleam and somehow ended up here, I can recommend you\n&gt; to first check out the\n&gt; [Gleam&#39;s Exercism track](https://exercism.org/tracks/gleam/concepts).\n&gt; It&#39;s a great resource to get started!\n\nFirst things first, you can import the Glam package in your project&#39;s\nmodules like this:\n\n```gleam\nimport glam/doc.{type Document}\n```\n\n## Basic building blocks\n\nLet&#39;s start by writing the type of the function that can turn lists into pretty\nprinted lists; after all it is always a good idea to write down the types and\nlet them guide you:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  todo as &quot;Write an amazing pretty printer&quot;\n}\n```\n\nThe first thing you can notice is that the pretty printing function&#39;s return\ntype is `Document`. But wait, weren&#39;t we supposed to pretty print the list?\nWhy isn&#39;t the function returning a `String`?\n\nHere is a crucial point of how the Glam package handles pretty printing:\nyou can _describe_ the structure of the text to pretty print with some basic\nbuilding blocks, and Glam will do the heavy lifting of trying to find the\nbest layout for you.\nAs you may have guessed those building blocks are `Document`s.\n\nAs long as you can turn something into a `Document`, Glam will pretty print it\nfor you. Quite neat, isn&#39;t it?\n\n### A naive approach\n\nFirst of all, we need a way to create a `Document` from a `String`,\nluckily Glam has got us covered with the `doc.from_string` function which does\nexactly that:\n\n```gleam\ndoc.from_string(&quot;Hello, world!&quot;)\n|&gt; doc.to_string(80)\n// -&gt; Hello, world!\n```\n\nThere are a few things to unpack in this example:\n\n- `doc.from_string` takes a string and turns it into a `Document`\n- `doc.to_string` does the opposite: given a maximum line width and a document,\n  it turns it into a pretty printed string. Later we will discover how the\n  maximum line width can influence the pretty printing process\n\nWith `doc.from_string` in our toolbox we can now turn each of the items of the\nlist into documents:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let docs = list.map(list, list_item_to_document)\n\n  todo as &quot;Still not enough...&quot;\n}\n```\n\n`docs` is a list of documents but in order to be printed with `doc.to_string`\nit needs to be turned into a single `Document`.\nTo do that we can use the `doc.concat` function which takes a list of documents\nas input and joins them together into a single document:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(item) }\n  let docs = list.map(list, list_item_to_document)\n  docs |&gt; doc.concat\n}\n```\n\nWe&#39;re now ready to try `pretty_list`:\n\n```gleam\n[&quot;Hello&quot;, &quot;world!&quot;]\n|&gt; pretty_list\n|&gt; doc.to_string(80)\n// -&gt; &quot;Hello&quot;&quot;world!&quot;\n```\n\nNot quite list-looking but we&#39;re getting there... adding commas and brackets can\nbe done with some standard library functions on lists:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n\n  list.map(list, list_item_to_document)\n  |&gt; list.intersperse(doc.from_string(&quot;, &quot;)) // put a comma between each doc\n  |&gt; list.prepend(doc.from_string(&quot;[&quot;))      // starting open bracket\n  |&gt; list.append([doc.from_string(&quot;]&quot;)])     // final closing bracket\n  |&gt; doc.concat                              // join everything together\n}\n```\n\n```gleam\n[&quot;Hello&quot;, &quot;world!&quot;]\n|&gt; pretty_list\n|&gt; doc.to_string(80)\n// -&gt; [&quot;Hello&quot;, &quot;world!&quot;]\n```\n\nAwesome, with a couple of lines of code you can now pretty print a list of\nstrings.\nLet&#39;s play a bit more with the maximum line width and see how the list gets\nprinted:\n\n```gleam\npub fn main() {\n  let list = [&quot;Hello&quot;, &quot;world!&quot;]\n  pretty_list(list)\n  |&gt; doc.to_string(5) // &lt;- this is really small and the list\n  |&gt; io.println       //    should be printed on multiple lines\n}\n\n// main()\n// -&gt; [&quot;Hello&quot;, &quot;world!&quot;]\n```\n\nWe would expect the list to be split on multiple lines like in the example I\nshowed you at the beginning of the tutorial.\nIt looks like `doc.concat` is not enough to create beautiful documents.\nIn the next section we&#39;ll discover more building blocks that can get us closer\nto a pretty printed list.\n\n### A less naive approach\n\nTo understand the problem here we need to get into a bit more detail on how the\npretty printing of documents works:\n\n- When the pretty printer has to display a `doc.from_string` it simply outputs\n  the string passed to it as an argument leaving it unchanged\n- When the pretty printer has to display a list of documents joined with\n  `doc.concat`, it formats each document one and joins them together without\n  adding any whitespace or trying to split them to fit on a line\n\n&gt; If you already have some experience in Gleam and functional programming\n&gt; you may have noticed that a pretty printer that only has these two primitives\n&gt; is just a more convoluted way of concatenating strings, `doc.concat` would act\n&gt; as a `list.fold(strings, from: &quot;&quot;, with: fn(acc, string) { acc &lt;&gt; string })`\n&gt;\n&gt; If this `fold` thing looks like gibberish to you, don&#39;t worry!\n&gt; You can look at the\n&gt; [Gleam&#39;s Exercism track](https://exercism.org/tracks/gleam/concepts), it&#39;s a\n&gt; great way to get started and hone your functional programming skills! (At\n&gt; this point I&#39;m sure you get how much I love Gleam&#39;s Exercism track)\n\n#### Grouping docs\n\nWhat we really need is a way to explicitly tell the pretty printer it is\nallowed to break a document if it does not fit on a single line.\n\nThis is the job for `doc.group` and `doc.space`:\n\n- When the pretty printer runs into a `doc.group` it first tries to render the\n  document on a single line, just like `doc.concat`. The catch is that if the\n  document would not fit on a single line, the pretty printer will break all the\n  spaces inside it\n- A `doc.space` is a special kind of document that gets normally rendered as a\n  single whitespace: `&quot; &quot;`. However, the pretty printer can turn these spaces\n  into newlines in order to split a `doc.group` that would not fit on a single\n  line\n\nLet&#39;s look at an example to make things clearer:\n\n```gleam\nlet doc =\n  [doc.from_string(&quot;Hello&quot;), doc.space, doc.from_string(&quot;world!&quot;)]\n  |&gt; doc.concat\n  |&gt; doc.group\n\ndoc.to_string(doc, 80)\n// -&gt; Hello world!\ndoc.to_string(doc, 10)\n// -&gt;\n// Hello\n// world!\n```\n\nAs you can see, in the second example, where the line width wouldn&#39;t allow\n`&quot;Hello&quot;` and `&quot;world!&quot;` to fit onto a single line, the space is rendered as a\nnewline.\n\n#### Even more examples\n\nBefore going back to our list pretty printer we should experiment a little bit\nmore with `doc.group` and `doc.space`.\n\nIn particular, there&#39;s a little detail I omitted in my first explanation of how\ngroup formatting works. Let&#39;s see how the pretty printer deals with nested\ngroups:\n\n```gleam\nlet food =\n  [&quot;lasagna&quot;, &quot;ravioli&quot;, &quot;pizza&quot;]\n  |&gt; list.map(doc.from_string)\n  |&gt; list.intersperse(with: doc.space)\n  |&gt; doc.concat\n  |&gt; doc.group\n\nlet food_i_love =\n  // Sorry for being stereotypically italian\n  [doc.from_string(&quot;Food I love:&quot;), doc.space, food]\n  |&gt; doc.concat\n  |&gt; doc.group\n```\n\n&gt; You&#39;ll find out that we frequently need to do something like this to join\n&gt; documents together using a custom separator:\n&gt;\n&gt; ```gleam\n&gt; list_of_docs\n&gt; |&gt; list.intersperse(with: some_separator)\n&gt; |&gt; doc.concat\n&gt; ```\n&gt;\n&gt; To make things a bit easier to read you can also use the `doc.join` function\n&gt; which is a shorthand for an `intersperse` followed by `concat`:\n&gt;\n&gt; ```gleam\n&gt; list_of_docs\n&gt; |&gt; doc.join(with: some_separator)\n&gt; ```\n\nHere we have two groups, one inside the other, let&#39;s play around with the\nmaximum line width and see how this gets printed:\n\n```gleam\nfood_i_love |&gt; doc.to_string(80)\n// -&gt; Food I love: lasagna ravioli pizza\n\nfood_i_love |&gt; doc.to_string(20)\n// -&gt;\n// Food I love:\n// lasagna\n// ravioli\n// pizza\n\nfood_i_love |&gt; doc.to_string(30)\n// -&gt;\n// Food I love:\n// lasagna ravioli pizza\n```\n\n- In the first example we gave the document enough space to fit on a single\n  line, all spaces get rendered as a normal whitespace and everything ends up on\n  a single line\n- In the second example the line width is quite small: neither the outer nor\n  the inner document could fit on a single line so all spaces become newlines\n- The most interesting example is the third one where the outer group is split\n  while the inner is not.\n  What ends up happening in the pretty process could be described like this:\n  - The pretty printer first tries to render everything on a single line but\n    the resulting document would exceed the maximum width of 30 chars\n  - So it decides to split the outer group and its space gets rendered as a\n    newline\n  - On the new line the pretty printer tries to render the inner document\n    and it fits just right! So its spaces, contrary to what happened to those\n    of the outer group, are not turned into newlines\n\nSo when the pretty printer ends up breaking a group it may still not break the\ninner groups it&#39;s composed of. Each subgroup is considered _separately_.\n\n&gt; It can get some time to wrap your head around how groups can influence the\n&gt; pretty printing process.\n&gt; If you want to test your understanding of `doc.group`, try to look at this\n&gt; example and guess how it will get printed at different line widths:\n&gt;\n&gt; ```gleam\n&gt; let abcd =\n&gt;   [&quot;a&quot;, &quot;b&quot;, &quot;c&quot;, &quot;d&quot;]\n&gt;   |&gt; list.map(doc.from_string)\n&gt;   |&gt; doc.join(with: doc.space)\n&gt;   |&gt; doc.group\n&gt;\n&gt; let efgh =\n&gt;   [&quot;e&quot;, &quot;f&quot;, &quot;g&quot;, &quot;h&quot;]\n&gt;   |&gt; list.map(doc.from_string)\n&gt;   |&gt; doc.join(with: doc.space)\n&gt;   |&gt; doc.group\n&gt;\n&gt; let doc =\n&gt;   [doc.from_string(&quot;some letters:&quot;), abcd, efgh]\n&gt;   |&gt; doc.join(with: doc.space)\n&gt;   |&gt; doc.group\n&gt;\n&gt; // Try different widths and guess what the document will look like\n&gt; doc |&gt; doc.to_string(...) \n&gt; ```\n\n#### Back onto lists\n\nIt shouldn&#39;t be too hard to update the previous iteration of `pretty_list` to\ntake advantage of groups and breakable spaces. Let&#39;s try it out:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.space])\n  let close_square = doc.concat([doc.space, doc.from_string(&quot;]&quot;)])\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n  // ^-- Don&#39;t forget the group, otherwise the pretty\n  //     printer won&#39;t be able to split on spaces\n}\n```\n\nThe main difference is that the comma that we use as a separator is now followed\nby a `doc.space` which the pretty printer can break if the list is too long and\ndoes now fit on a single line. The same goes for square brackets: the open\nbracket is followed by a `doc.space` so that, if the document gets broken, the\nlist&#39;s first item will end up on a new line. Likewise, the closed bracket is\npreceded by a space so that it can end up on its own line:\n\n```gleam\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]) |&gt; doc.to_string(80)\n// -&gt; [ &quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot; ]\n\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]) |&gt; doc.to_string(10)\n// -&gt;\n// [\n// &quot;Gleam&quot;,\n// &quot;is&quot;,\n// &quot;fun&quot;  \n// ]\n```\n\nWe&#39;re getting there! There are just a couple of things we need to fix before\ncalling it a day:\n\n- The list items should be indented by two spaces\n- When the list is rendered on multiple lines, the last item should have a\n  trailing comma\n- When the list is rendered on a single line we do not want the first and last\n  items to be separated with a space from the brackets\n\n## Nesting things\n\nTo solve the first problem, we need to add some nesting whitespace whenever\na space is turned into a newline. In order to do so we can use the `doc.nest`\nfunction:\n\n```gleam\ndoc.nest(my_document, by: 2)\n```\n\nThis tells the pretty printer that, whenever it renders a space as a newline\nwhen formatting `my_document` it should also indent it by two spaces.\nLet&#39;s get back to the `food_i_love` example and see how the rendering changes\nby adding a nesting of two spaces:\n\n```gleam\nfood_i_love\n|&gt; doc.nest(by: 2)\n|&gt; doc.to_string(20)\n\n// -&gt;\n// Food I love:\n//   lasagna\n//   ravioli\n//   pizza\n```\n\nNow we can add nesting to the `pretty_list` document:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.space])\n  let close_square = doc.concat([doc.space, doc.from_string(&quot;]&quot;)])\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.nest(by: 2)\n  // ^-- every space before the final bracket will get indented\n  //     by two spaces if turned into a new line\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n}\n```\n\nNotice how the indentation applies to all the spaces _except_ the one that comes\nbefore the closed square bracket. This way, the closed square bracket will not\nbe indented and kept on the same nesting level as the corresponding open\nbracket:\n\n```gleam\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]) |&gt; doc.to_string(10)\n// -&gt;\n// [\n//   &quot;Gleam&quot;,\n//   &quot;is&quot;,\n//   &quot;fun&quot;\n// ]\n```\n\n## Custom breaks\n\nThe last piece of work we need to do to get the desired look for a pretty\nprinted list is adding a trailing comma and removing whitespace before and\nafter the square brackets:\n\n```gleam\n// This:\n[&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]\n\n// not this:\n[ &quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot; ]\n```\n\n```gleam\n// This:\n[\n  &quot;Gleam&quot;,\n  &quot;is&quot;,\n  &quot;fun!&quot;,\n]\n\n// not this:\n[\n  &quot;Gleam&quot;,\n  &quot;is&quot;,\n  &quot;fun!&quot;\n]\n```\n\nBoth problems can be solved by introducing a new powerful document constructor:\n`doc.break`.\n\nUp until now `doc.space` was the only document that the pretty printer could\nturn into a newline when splitting a group. Turns out that `doc.break` is a\nmore general version of `doc.space`. Here is how it can be called:\n\n```gleam\ndoc.break(&quot;unbroken&quot;, &quot;broken&quot;)\n```\n\n- Its first argument is the string that will be displayed if the pretty\n  printer decides not to break the document\n- Its second argument is the string that will be displayed if the pretty\n  printer decides to break the document _before_ inserting a newline\n\n### The pretty printing algorithm\n\nThis description of `doc.break` may sound a bit hand-wavy so, to clear things\nup, let&#39;s have a final look at how the formatting of documents actually works:\n\n- The pretty printer always tries to put a group onto a single line by treating\n  every `doc.break` as if it were its first argument\n- If the group exceeds the maximum line width then all its `break`s are rendered\n  as their second argument and immediately followed by a newline\n  - If the document is also nested, the pretty printer adds a nesting of the\n    given level after the newline\n- After breaking the outermost group, the pretty printer considers every nested\n group on its own, deciding once again if it can fit on the new line or not\n\nLet&#39;s look at an example:\n\n```gleam\nlet visible_space = doc.break(&quot;•&quot;, &quot;↩&quot;)\nlet doc =\n  [&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]\n  |&gt; doc.join(with: visible_space)\n  |&gt; dog.group\n\ndoc |&gt; doc.to_string(80)\n// -&gt; Gleam•is•fun!\n\ndoc |&gt; doc.to_string(10)\n// -&gt;\n// Gleam↩\n// is↩\n// fun\n```\n\nAs you can see, when the pretty printer decides to split the document, all of\nits breaks are displayed as `&quot;↩&quot;`; if it is not broken they are displayed as\n`&quot;•&quot;`.\n\n&gt; `doc.space` can actually be defined in terms of `doc.break` (and that&#39;s how\n&gt; it is defined in Glam!) as `doc.break(&quot; &quot;, &quot;&quot;)`:\n&gt;\n&gt; - If the group is not broken we render a single whitespace\n&gt; - If the group gets broken we do not render anything and let the pretty\n&gt;   printer do its magic and add a newline after each space\n\n### Back to lists one last time\n\nThanks to `doc.break` we can solve both our remaining problems: first of all we\nwant to display a trailing comma after the last element _only if_ the list is\nsplit onto multiple lines:\n\n```gleam\nlet trailing_comma = break(&quot;&quot;, &quot;,&quot;)\n```\n\nThis comma will only be displayed if the group it belongs to is broken by the\npretty printer:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.space])\n  let trailing_comma = break(&quot;&quot;, &quot;,&quot;)\n  let close_square = doc.concat([trailing_comma, doc.from_string(&quot;]&quot;)])\n  // ^-- we add the trailing comma right before the closing bracket\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n}\n```\n\nAs a finishing touch we need to remove the whitespace that gets printed after\nthe open bracket. The perfect fit for this job is once again `doc.break`:\n\n```gleam\nlet soft_break = doc.break(&quot;&quot;, &quot;&quot;)\n```\n\n`soft_break` will never be rendered, its only purpose is to act as a possible\nbreaking point for the pretty printer to split a group on multiple lines.\n\nSince `soft_break` can be useful in a lot of different settings, the Glam\npackage already exposes it as a constant so that you do not have to redefine it\nevery time you need it:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.soft_break])\n  // ^-- so that there won&#39;t be a whitespace after the open bracket\n  //     but the pretty printer can still split it\n  let trailing_comma = doc.break(&quot;&quot;, &quot;,&quot;)\n  let close_square = doc.concat([trailing_comma, doc.from_string(&quot;]&quot;)])\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n}\n```\n\nAnd that&#39;s it, the pretty printer is complete!\n\n## Recap\n\nPhew! That was quite a long tutorial so give yourself a pat on the back and\nenjoy the beautifully pretty printed lists.\n\nWe covered most of the Glam API and got an understanding of how each piece can\ninfluence the pretty printing process:\n\n- `doc.from_string` to generate documents from strings\n- `doc.concat` to join together a bunch of documents\n- `doc.group` and `doc.space` to allow the pretty printer to split documents\n- `doc.nest` to nest documents when the pretty printer decides to split them\n- `doc.break` to conditionally render strings depending if their group gets\n  split or not\n\nWe&#39;ve also used some utility methods like:\n\n- `doc.join` which is equivalent to interspersing and then using `doc.concat`\n- `doc.prepend` to prepend a document to an existing one\n- `doc.append` to append a document to an existing one\n\n## What to do next?\n\nIf you still feel hungry for pretty printing content, I&#39;ve got you covered:\nhave a look at Glam&#39;s\n[learning materials page](https://hexdocs.pm/glam/learning_materials.html).\nThere you&#39;ll find more step-by-step tutorials that will guide you through the\nimplementation of cool pretty printers.\n","url":"pretty_printing_glamorous_lists.html"},{"doc":"glam","title":"glam","content":"# Pretty printing TODO lists\n\nThis tutorial assumes that you already have some familiarity with\npretty printing in Glam and with its API.\n\nIf you feel like you need to brush up on your pretty printing skills, you can\nfirst go through the\n[introductory tutorial](https://hexdocs.pm/glam/pretty_printing_glamorous_lists.html)\nand come back to this one.\n\n## Goals\n\nThis tutorial will show you how to write a pretty printer for TODO lists.\nWhile doing so, you&#39;ll learn some of the remaining bits of Glam&#39;s API that\nwere not covered in the introductory tutorial; you&#39;ll also get more familiar\nwith how nesting works.\n\n## TODO lists\n\nFirst of all, let&#39;s have a look at the data structure we&#39;re going to\npretty print:\n\n```gleam\ntype TodoList =\n  List(Task)\n\ntype Task {\n  Task(status: Status, description: String, subtasks: TodoList)\n}\n\ntype Status {\n  Todo\n  Done\n  InProgress\n}\n```\n\nA `TodoList` is a list of `Task`s; each task has a description, a status and\nmany subtasks.\nLet&#39;s look at an example of such data structure _(any similarity to actual_\n_persons ~~me~~ or actual events ~~me writing this tutorial~~ is purely_\n_coincidental)_:\n\n```gleam\nlet todo_list = [\n  Task(InProgress, &quot;publish Glam v1.1.0&quot;, [\n    Task(InProgress, &quot;write a tutorial on todo lists&quot;, []),\n    Task(InProgress, &quot;add `doc.flex_break`&quot;, [\n      Task(Done, &quot;add the appropriate type variant&quot;, []),\n      Task(Done, &quot;implement the missing cases&quot;, []),\n      Task(Todo, &quot;add some tests&quot;, []),\n    ]),\n  ]),\n  Task(Todo, &quot;get some sleep&quot;, []),\n]\n```\n\nWe want each list item to be on its own line and to be preceded by a bullet\npoint. Also, each list of subtasks should be indented by two spaces more than\nits parent.\n\nThe TODO list I&#39;ve just shown you would look like this:\n\n```text\n- […] publish Glam v1.1.0\n  - […] write a tutorial on todo lists\n  - […] add `doc.flex_break`\n    - [X] add the appropriate type variant\n    - [X] implement the missing cases\n    - [ ] add some tests\n- [ ] get some sleep\n```\n\n## A mostly wrong implementation\n\nWe need to find a way to turn a `TodoList` into a `Document` so that Glam will\nbe able to pretty print it for us. The function we&#39;re looking for looks like\nthis:\n\n```gleam\nfn tasks_to_doc(tasks: TodoList) -&gt; Document {\n  todo as &quot;Write an amazing pretty printer&quot;\n}\n```\n\nWe can turn this problem into something more approachable by first tackling the\npretty printing of a single task:\n\n```gleam\nfn task_to_doc(task: Task) -&gt; Document {\n  todo as &quot;Another amazing pretty printer&quot;\n}\n```\n\nWith this definition it&#39;s quite straightforward to implement a\n_mostly_ correct (more on this later) `tasks_to_doc`:\n\n```gleam\nfn tasks_to_doc(tasks: TodoList) -&gt; Document {\n  list.map(tasks, task_to_doc)\n  |&gt; doc.join(with: doc.soft_break)\n}\n```\n\n### Pretty printing a task\n\nIn order to pretty print a single task, we first have to create the appropriate\nbullet point according to its status:\n\n```gleam\nfn status_to_bullet(status: Status) -&gt; String {\n  case status {\n    Todo -&gt; &quot;- [ ]&quot;\n    Done -&gt; &quot;- [X]&quot;\n    InProgress -&gt; &quot;- […]&quot;\n  }\n}\n```\n\nNow we&#39;re ready to turn a task into a document:\n\n```gleam\nfn task_to_doc(task: Task) -&gt; Document {\n  let task_line = status_to_bullet(task.status) &lt;&gt; &quot; &quot; &lt;&gt; task.description\n  let task_doc = doc.from_string(task_line)\n  \n  todo as &quot;Bear with me a little longer...&quot;\n}\n```\n\nThe document for the task we&#39;ve just created is obtained from a single string:\nif we joined the pieces with a `doc.space`, the pretty printer could split the\nbullet point and the task description on different lines.\nWhat we want is to have the task always on the same line as its bullet.\n\nNow we can turn our attention to the list of subtasks; we can handle those with\na mutually recursive call to `tasks_to_doc`:\n\n```gleam\nfn task_to_doc(task: Task) -&gt; Document {\n  let task_line = status_to_bullet(task.status) &lt;&gt; &quot; &quot; &lt;&gt; task.description\n  let task_doc = doc.from_string(task_line)\n\n  case list.is_empty(tasks.subtasks) {\n    True -&gt; task_doc\n    False -&gt;\n      [task_doc, doc.soft_break, docs_to_task(task.subtasks)]\n      |&gt; doc.concat\n  }\n}\n```\n\nWe just have to be careful and remember to not add any space if there&#39;s actually\nno subtask to display.\nWhen there are subtasks, though, we join their document and the task&#39;s one with\na `doc.soft_break`.\n\nLet&#39;s try this out and see how a TODO list is pretty printed:\n\n```gleam\n[\n  Task(Todo, &quot;groceries&quot;, [\n    Task(Todo, &quot;lentils&quot;, [])\n    Task(Todo, &quot;carrots&quot;, [])\n  ])\n]\n|&gt; tasks_to_doc\n|&gt; doc.to_string(10_000)\n// -&gt; - [ ] groceries- [ ] lentils- [ ] carrots\n```\n\nLooks like everything ended up on a single line. What we wanted was to get a\ntask per line, no matter how wide it is.\n\nThe problem is that, when the pretty printer has to deal with a `doc.soft_break`\n(or any other kind of `break`), it only splits it if the group it belongs to\ndoesn&#39;t fit on a single line.\nIf, like in this example, we provide the pretty printer enough space, it will\ngladly keep everything on a single line!\n\nThere is a quick and dirty solution: just trick the pretty printer and use `0`\nas the maximum width.\nThen, the pretty printer will always split every single group it runs into:\n\n```gleam\n[\n  Task(Todo, &quot;first&quot;, []),\n  Task(Todo, &quot;second&quot;, []),\n]\n|&gt; tasks_to_doc\n|&gt; doc.to_string(0)\n// -&gt;\n// - [ ] first\n// - [ ] second\n```\n\nWhile this works for such a simple example, the problem with this approach is\nthat it breaks down quite easily for more complex documents: imagine you had\nonly a portion of the document that you wanted to always break; by setting the\nline width to `0` you&#39;re going to always break _every single group_.\n\n## Forcing a break\n\nThere&#39;s a better way to force the pretty printer to always break a given break:\n`doc.force_break`.\nThis function takes a document as input and forces the pretty printer to break\nthe `doc.break` it is made of:\n\n```gleam\nlet example = \n  [&quot;first&quot;, &quot;second&quot;]\n  |&gt; list.map(doc.from_string)\n  |&gt; doc.join(with: doc.space)\n  |&gt; doc.force_break\n\ndoc.to_string(example, 10_000)\n// -&gt;\n// first\n// second\n```\n\nAs you can see, despite having plenty of room to fit both `&quot;first&quot;` and\n`&quot;second&quot;` on a single line, the pretty printer was forced to break the space.\n\nThere&#39;s a detail to pay attention to: `doc.force_break` _does not work on_\n_groups_! The pretty printer will always try to first fit a group onto a single\nline, ignoring any `doc.force_break` wrapping it:\n\n```gleam\nlet example =\n  [&quot;first&quot;, &quot;second&quot;]\n  |&gt; list.map(doc.from_string)\n  |&gt; doc.join(with: doc.space)\n  |&gt; doc.group\n  |&gt; doc.force_break\n\ndoc.to_string(example, 10_000)\n// -&gt; first second\n```\n\nSimply wrapping everything in a `doc.group` _before_ calling `doc.force_break`\nrenders it useless: the pretty printer will just ignore it and treat the group\nas it usually would.\n\n### Fixing the TODO list pretty printer\n\nWe can take advantage of `doc.force_break` to ask the pretty printer to split\nthe items on newlines, no matter the line width.\nThe required change to get a correct implementation of `tasks_to_doc` is minimal:\n\n```gleam\nfn tasks_to_doc(tasks: TodoList) -&gt; Document {\n  list.map(tasks, task_to_doc)\n  |&gt; doc.join(with: doc.soft_break)\n  |&gt; doc.force_break\n}\n```\n\nSimply wrapping the resulting `Document` in a `doc.force_break` is enough to\nmake sure the pretty printer always ends up splitting the `doc.soft_break` we\nused to join the tasks.\n\nLet&#39;s have a look at the pretty printed list now:\n\n```gleam\n[\n  Task(Todo, &quot;groceries&quot;, [\n    Task(Todo, &quot;lentils&quot;, [])\n    Task(Todo, &quot;carrots&quot;, [])\n  ])\n]\n|&gt; tasks_to_doc\n|&gt; doc.to_string(10_000)\n// -&gt;\n// - [ ] groceries\n// - [ ] lentils\n// - [ ] carrots\n```\n\nAlmost perfect! We just need to indent the subtasks to make this look as we\nwanted.\n\n## Nesting subtasks\n\nAs we&#39;ve already seen in the introductory tutorial, we can use `doc.nest` to\nincrease the padding added after each newline inserted by the pretty printer.\n\nOnce again the code change to make our function work as intended is minimal:\n\n```gleam\nfn task_to_doc(task: Task) -&gt; Document {\n  let task_line = status_to_bullet(task.status) &lt;&gt; &quot; &quot; &lt;&gt; task.description\n  let task_doc = doc.from_string(task_line)\n\n  case list.is_empty(tasks.subtasks) {\n    True -&gt; task_doc\n    False -&gt;\n      [task_doc, doc.soft_break, docs_to_task(task.subtasks)]\n      |&gt; doc.concat\n      |&gt; doc.nest(by: 2)\n      // ^-- we just needed to add this single line\n  }\n}\n```\n\nOne nice thing about `doc.nest` is that it _increases_ the nesting level of the\ndocument. This means that, any other nesting that comes before it is\nautomatically increased by the same amount.\nThat&#39;s why this small change is enough to deal with deeply nested task lists\nand each sublist is going to be nested at the right level.\n\nIf we now call `tasks_to_doc` on the TODO list I showed you at the beginning of\nthe tutorial what we get is this nice-looking list:\n\n```gleam\ntasks_to_doc(todo_list)\n|&gt; doc.to_string(10_000)\n// -&gt;\n// - […] publish Glam v1.1.0\n//   - […] write a tutorial on todo lists\n//   - […] add `doc.flex_break`\n//     - [X] add the appropriate type variant\n//     - [X] implement the missing cases\n//     - [ ] add some tests\n// - [ ] get some sleep\n```\n\n## Recap\n\nOur final implementation is just a bit less than 30 lines of pretty printing\nbut we got quite a nice value out of it:\n\n- You learned about `doc.force_break` as a way to force the pretty printer\n  to nicely handle break that should always be split\n- Hopefully you got a better understanding of how `doc.nest` can help you\n  nest arbitrarily deep data structures\n\nIf you want to take a look at the full implementation code, you can find it\n[here](https://github.com/giacomocavalieri/glam/blob/main/src/examples/todo_lists.gleam).\n\n## What to do next?\n\nIf you want to test your skills on a real-world example, you can have a look at\nthe\n[JSON pretty printing tutorial](https://hexdocs.pm/glam/pretty_printing_JSON.html),\nI highly recommend you try it: it&#39;s really rewarding and you&#39;ll never look at\nonline JSON formatters in the same way!\n\nThere&#39;s also another tutorial on\n[pretty printing lovely error messages](https://hexdocs.pm/glam/pretty_printing_lovely_error_messages.html)\nyou can look at.\nIt&#39;s one of the tutorials I had the most fun writing, I hope you find it\ninteresting as well!\n","url":"pretty_printing_todo_lists.html"},{"doc":"glam","title":"glam","content":"# Pretty printing JSON\n\nThis tutorial assumes that you already have some familiarity with\npretty printing in Glam and with its API.\n\nIf you feel like you need to brush up on your pretty printing skills, you can\nfirst go through the\n[introductory tutorial](https://hexdocs.pm/glam/pretty_printing_glamorous_lists.html) and come\nback to this one.\n\n## Goals\n\nThis tutorial will walk you through the implementation of a pretty printer for\nJSON data. It is meant to be a detailed description of the implementation of a\nreal-life pretty printer, a bit more complicated than those shown in previous\ntutorials.\n\nThis tutorial assumes you already have some knowledge of the Glam API and won&#39;t\ndwell on explaining _how_ the pretty printing works, if you feel confused it may\nbe best to first read the\n[introductory tutorial](https://hexdocs.pm/glam/pretty_printing_glamorous_lists.html).\n\n## What&#39;s in a JSON?\n\nFirst things first, let&#39;s have a look at the data structure we&#39;re going to use\nto represent JSON in Gleam:\n\n```gleam\ntype JSON {\n  String(String)\n  Number(Float)\n  Bool(Bool)\n  Null\n  Array(List(JSON))\n  Object(List(#(String, JSON)))\n}\n```\n\nA valid JSON object is either a primitive type (a number, a string, a bool or\nthe literal `null`), or a collection (a sequence of JSON objects, or a\ncollection of key/value pairs).\n\n## A running example\n\nMe and my friends are hosting a book club and I&#39;m keeping a record for each book\nwe end up reading and our ratings.\nAn example record, as written with the JSON type defined above, looks like this:\n\n```gleam\nlet the_sundial =\n  Object([\n    #(&quot;title&quot;, String(&quot;The sundial&quot;)),\n    #(&quot;author&quot;, String(&quot;Shirley Jackson&quot;)),\n    #(&quot;publication_year&quot;, Number(1958.0)),\n    #(&quot;read&quot;, Bool(True)),\n    #(&quot;characters&quot;, Array([\n      String(&quot;Mrs. Halloran&quot;),\n      String(&quot;Essex&quot;),\n    ])),\n    #(&quot;average_rating&quot;, Number(5.0)),\n    #(&quot;ratings&quot;, Array([\n      Object([\n        #(&quot;from&quot;, String(&quot;Ben&quot;)),\n        #(&quot;value&quot;, Number(5.0)),\n      ]),\n      Object([\n        #(&quot;from&quot;, String(&quot;Giacomo&quot;)),\n        #(&quot;value&quot;, Number(5.0)),\n      ]),\n    ])),\n  ])\n```\n\nI want to write a pretty printer for JSON data to display this in a nice,\nhuman-readable way. The desired output for this object would look like this:\n\n```json\n{\n  title: &quot;The sundial&quot;,\n  author: &quot;Shirley Jackson&quot;,\n  publication_year: 1958.0,\n  read: true,\n  characters: [ &quot;Mrs. Halloran&quot;, &quot;Essex&quot; ],\n  average_rating: 5.0,\n  ratings: [\n    { from: &quot;Ben&quot;, value: 5.0 },\n    { from: &quot;Giacomo&quot;, value: 5.0 }\n  ]\n}\n```\n\nThis example already shows all the rules we want our pretty printer to follow,\nlet&#39;s write them down as a reference:\n\n- All strings must be wrapped in double quotes `&quot;`\n- Numbers are always printed with their decimal part, even when it&#39;s 0\n- Booleans are printed as `true` and `false`\n- Null is printed as `null`\n- It should always try to keep an array on a single line, otherwise (like in\n  the case of `reviews`) it should be split with each item on a line of its own\n  and indented by two spaces\n- It should always try to keep an object on a single line (like the ratings),\n  otherwise, it should be split with each field/value pair on a line of its own\n- When displaying an object, the field name and the relative value should never\n  end up on different lines. If the value is an object/array that needs to be\n  split, its open parentheses must be kept on the same line as the field name\n  (for example look at the `ratings` field)\n- Neither objects nor arrays should have a trailing comma since it&#39;s not allowed\n  in JSON\n- When objects and arrays are displayed on a single line there should be a\n  single whitespace separating the parentheses and their items\n\nThat&#39;s quite a lot of constraints! But, as you&#39;ll see, it won&#39;t be too\ncomplicated to get a working pretty printer.\n\n## Dealing with primitive data\n\nIn order to do anything with the `JSON` type we must first pattern match on it:\n\n```gleam\nfn json_to_doc(json: JSON) -&gt; Document {\n  case json {\n    String(string) -&gt; todo as &quot;string&quot;\n    Number(number) -&gt; todo as &quot;number&quot;\n    Bool(bool) -&gt; todo as &quot;bool&quot;\n    Null -&gt; todo as &quot;null&quot;\n    Array(objects) -&gt; todo as &quot;array&quot;\n    Object(fields) -&gt; todo as &quot;object&quot;\n  }\n}\n```\n\nDealing with the primitive data types like string, null, number and bool\nshouldn&#39;t be too surprising:\n\n```gleam\nfn json_to_doc(json: JSON) -&gt; Document {\n  case json {\n    String(string) -&gt; doc.from_string(&quot;\\&quot;&quot; &lt;&gt; string &lt;&gt; &quot;\\&quot;&quot;)\n    Number(number) -&gt; doc.from_string(float.to_string(number))\n    Bool(bool) -&gt; bool_to_doc(bool)\n    Null -&gt; doc.from_string(&quot;null&quot;)\n    Array(objects) -&gt; todo as &quot;array&quot;\n    Object(fields) -&gt; todo as &quot;object&quot;\n  }\n}\n```\n\nTo turn a boolean value into a document we can define a helper function\n`bool_to_doc` (you could also transform the value directly inside the pattern\nmatching branch, but I&#39;d rather define an helper function to keep things\ntidy):\n\n```gleam\nfn bool_to_doc(bool: Bool) -&gt; Document {\n  bool.to_string(bool)\n  |&gt; string.lowercase\n  |&gt; doc.from_string\n}\n```\n\nThis is already enough to print simple JSON objects:\n\n```gleam\nBool(True) |&gt; json_to_doc |&gt; doc.to_string(80)\n// -&gt; true\n\nNull |&gt; json_to_doc |&gt; doc.to_string(80)\n// -&gt; null\n\nString(&quot;json!&quot;) |&gt; json_to_doc |&gt; doc.to_string(80)\n// -&gt; &quot;json!&quot;\n```\n\nNot too exciting, but a nice starting point. Let&#39;s move on to arrays.\n\n## Pretty printing arrays\n\nNow that we got the simple stuff out of the way we can focus on arrays and\nobjects. We will start with the former:\n\n```gleam\nfn json_to_doc(json: JSON) -&gt; Document {\n  case json {\n    String(string) -&gt; doc.from_string(&quot;\\&quot;&quot; &lt;&gt; string &lt;&gt; &quot;\\&quot;&quot;)\n    Number(number) -&gt; doc.from_string(float.to_string(number))\n    Bool(bool) -&gt; bool_to_doc(bool)\n    Null -&gt; doc.from_string(&quot;null&quot;)\n    Array(objects) -&gt; array_to_doc(objects)\n    Object(fields) -&gt; todo as &quot;object&quot;\n  }\n}\n\nfn array_to_doc(objects: List(JSON)) -&gt; Document {\n  todo as &quot;array_to_doc&quot;\n}\n```\n\nOnce again, I&#39;m defining a helper function to keep the pattern matching clean\nand readable. As a good starting point, we can turn the array items into a\ndocument and join those with a comma and a breakable space:\n\n```gleam\nfn comma() -&gt; Document {\n  doc.from_string(&quot;,&quot;)\n}\n\nfn array_to_doc(objects: List(JSON)) -&gt; Document {\n  list.map(objects, json_to_doc)\n  |&gt; doc.join(with: doc.concat([comma(), doc.space]))\n}\n```\n\n&gt; The Glam package also has a `doc.concat_join` function which is a shorthand\n&gt; for writing `doc.join(..., with: doc.concat(...))`.\n&gt;\n&gt; The code I just showed you can be rewritten like this:\n&gt;\n&gt; ```gleam\n&gt; list.map(objects, json_to_doc)\n&gt; |&gt; doc.concat_join(with: [comma(), doc.space])\n&gt; ```\n\nNow if we try to print an array we get:\n\n```gleam\nlet array =\n  Array([Null, Bool(True), String(&quot;json!&quot;), Number(5.0)])\n  |&gt; json_to_doc\n\ndoc.to_string(array, 80)\n// -&gt; null, true, &quot;json!&quot;, 5.0\n\ndoc.to_string(array, 10)\n// -&gt;\n// null,\n// true,\n// &quot;json!&quot;,\n// 5.0\n```\n\nThat&#39;s looking pretty good, we just have to deal with the square brackets and\nnesting. Let&#39;s first tackle the brackets problem:\n\n```gleam\nfn array_to_doc(objects: List(JSON)) -&gt; Document {\n  list.map(objects, json_to_doc)\n  |&gt; doc.concat_join(with: [comma(), doc.space])\n  |&gt; doc.prepend_docs([doc.from_string(&quot;[&quot;), doc.space])\n  |&gt; doc.append_docs([doc.space, doc.from_string(&quot;]&quot;)])\n  |&gt; doc.group\n}\n```\n\nA few things to notice here:\n\n- We use `doc.prepend_docs` and `doc.append_docs` which are just shorthands for\n  `doc.prepend(doc.concat(...))` and `doc.append(doc.concat(...))`\n- We add a breakable `doc.space` after the open bracket and before the closing\n  one. When the pretty printer outputs the array on a single line there will\n  be a space to separate the item from the brackets; however, if the array does\n  not fit on a single line, that space will be broken making sure that open and\n  closed brackets will be kept on a separate line\n- We wrap everything in a `doc.group`: this tells the pretty printer that it is\n  allowed to break the spaces inside it if it doesn&#39;t fit on a single line\n\nLast, we can add indentation for when the array gets broken on multiple lines:\n\n```gleam\nfn array_to_doc(objects: List(JSON)) -&gt; Document {\n  list.map(objects, json_to_doc)\n  |&gt; doc.concat_join(with: [comma(), doc.space])\n  |&gt; doc.prepend_docs([doc.from_string(&quot;[&quot;), doc.space])\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append_docs([doc.space, doc.from_string(&quot;]&quot;)])\n  |&gt; doc.group\n}\n```\n\n`doc.nest` must go _before_ the closing bracket; otherwise, even its preceding\nspace would get indented, resulting in the closing bracket being more indented\nthan the opening one.\n\nIf don&#39;t want to take my word for it and want to see it with your own eyes, try\nto move `doc.group` below the closed parentheses;\nan array split on multiple lines would look like this:\n\n```json\n[\n  true,\n  false\n  ] // &lt;- this is indented!\n```\n\n## Pretty printing objects\n\nWe need one last piece to complete the JSON pretty printer: a way to nicely\noutput objects. You&#39;ll find out that it&#39;s remarkably similar to pretty printing\narrays.\n\n```gleam\nfn json_to_doc(json: JSON) -&gt; Document {\n  case json {\n    String(string) -&gt; doc.from_string(&quot;\\&quot;&quot; &lt;&gt; string &lt;&gt; &quot;\\&quot;&quot;)\n    Number(number) -&gt; doc.from_string(float.to_string(number))\n    Bool(bool) -&gt; bool_to_doc(bool)\n    Null -&gt; doc.from_string(&quot;null&quot;)\n    Array(objects) -&gt; array_to_doc(objects)\n    Object(fields) -&gt; object_to_doc(fields)\n  }\n}\n\nfn object_to_doc(fields: List(#(String, JSON))) -&gt; Document {\n  todo as &quot;object_to_doc&quot;\n}\n```\n\nHow to turn a field/value pair into a document? We can first display the name, a\ncolon, and the value document that we can get with a recursive call to\n`json_to_doc`:\n\n```gleam\nfn colon() -&gt; Document {\n  doc.from_string(&quot;:&quot;)\n}\n\nfn field_to_doc(field: #(String, JSON)) -&gt; Document {\n  let #(name, value) = field\n  let name_doc = doc.from_string(name)\n  let value_doc = json_to_doc(value)\n  [name_doc, colon(), doc.from_string(&quot; &quot;), value_doc]\n  |&gt; doc.concat\n}\n```\n\nThe only point that needs extra attention here is the use of `doc.from_string(&quot; &quot;)`\nas a whitespace instead of `doc.space`. If you remember from the introductory\ntutorial (or the package documentation), a `doc.space` can always be broken by\nthe pretty printer into a new line; however, we said we want the field name to\nalways stay on the same line as its value.\n\nTo get this result we must use a single whitespace that doesn&#39;t also act as a\npossible breaking point.\nSince the pretty printer will never split documents obtained from a\n`doc.from_string`, we can use it with a single whitespace string like in the\ncode snippet above.\n\n### Putting everything together\n\nNow that we can turn a single field/value pair into a document we can get back\nto the problem of turning an entire object into a document:\n\n- Transform every field/value pair into a document\n- Join those together with a comma and a breakable space\n- Enclose everything in open and closed curly brackets\n- Nest the object fields\n\n```gleam\nfn object_to_doc(fields: List(#(String, JSON))) -&gt; Document {\n  list.map(fields, field_to_doc)\n  |&gt; doc.concat_join(with: [comma(), doc.space])\n  |&gt; doc.prepend_docs([doc.from_string(&quot;{&quot;), doc.space])\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append_docs([doc.space, doc.from_string(&quot;}&quot;)])\n  |&gt; doc.group\n}\n```\n\nThat&#39;s it, we can now print JSON objects:\n\n```gleam\nlet object =\n  Object([#(&quot;field1&quot;, Null),  #(&quot;field2&quot;, Bool(True))])\n  |&gt; json_to_doc\n\ndoc.to_string(object, 80)\n// -&gt; { field1: null, field2: true }\n\ndoc.to_string(object, 20)\n// -&gt;\n// {\n//   field1: null,\n//   field2: true   \n// }\n```\n\n## Some finishing touches\n\nThe tutorial could end here and we&#39;d have written a rad JSON pretty printer.\nBut there&#39;s a small piece of redundant code that is an eyesore to me, let&#39;s have\na look at the code for turning objects and arrays into nice-looking documents:\n\n```gleam\nfn array_to_doc(objects: List(JSON)) -&gt; Document {\n  list.map(objects, json_to_doc)\n  |&gt; doc.concat_join(with: [comma(), doc.space])\n  |&gt; doc.prepend_docs([doc.from_string(&quot;[&quot;), doc.space])\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append_docs([doc.space, doc.from_string(&quot;]&quot;)])\n  |&gt; doc.group\n}\n\nfn object_to_doc(fields: List(#(String, JSON))) -&gt; Document {\n  list.map(fields, field_to_doc)\n  |&gt; doc.concat_join(with: [comma(), doc.space])\n  |&gt; doc.prepend_docs([doc.from_string(&quot;{&quot;), doc.space])\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append_docs([doc.space, doc.from_string(&quot;}&quot;)])\n  |&gt; doc.group\n}\n```\n\nThat&#39;s almost exactly the same code! We&#39;ve duplicated the logic with which we\nparenthesise objects and arrays, let&#39;s pull it out in its own function:\n\n```gleam\nfn parenthesise(document: Document, open: String, close: String) -&gt; Document {\n  document\n  |&gt; doc.prepend_docs([doc.from_string(open), doc.space])\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append_docs([doc.space, doc.from_string(close)])\n  |&gt; doc.group\n}\n```\n\nAnd the original code can be rewritten using it:\n\n```gleam\nfn array_to_doc(objects: List(JSON)) -&gt; Document {\n  list.map(objects, json_to_doc)\n  |&gt; doc.concat_join(with: [comma(), doc.space])\n  |&gt; parenthesise(&quot;[&quot;, &quot;]&quot;)\n}\n\nfn object_to_doc(fields: List(#(String, JSON))) -&gt; Document {\n  list.map(fields, field_to_doc)\n  |&gt; doc.concat_join(with: [comma(), doc.space])\n  |&gt; parenthesise(&quot;{&quot;, &quot;}&quot;)\n}\n```\n\nNow I can proudly say we&#39;re done. If you try printing the big object I showed\nyou at the beginning of the tutorial, you&#39;ll see an output like this:\n\n```gleam\njson_to_doc(the_sundial) |&gt; doc.to_string(20)\n// -&gt;\n// {\n//   title: &quot;The sundial&quot;,\n//   author: &quot;Shirley Jackson&quot;,\n//   publication_year: 1958.0,\n//   read: true,\n//   characters: [\n//     &quot;Mrs. Halloran&quot;,\n//     &quot;Essex&quot;\n//   ],\n//   average_rating: 5.0,\n//   ratings: [\n//     {\n//       from: &quot;Ben&quot;,\n//       value: 5.0\n//     },\n//     {\n//       from: &quot;Giacomo&quot;,\n//       value: 5.0\n//     }\n//   ]\n// }\n```\n\nTry to play around with the line width to see how the output adapts to it!\n","url":"pretty_printing_JSON.html"},{"doc":"glam","title":"glam","content":"# Pretty printing lovely error messages\n\nThis tutorial assumes that you already have some familiarity with\npretty printing in Glam and with its API.\n\nIf you feel like you need to brush up on your pretty printing skills, you can\nfirst go through the\n[introductory tutorial](https://hexdocs.pm/glam/pretty_printing_glamorous_lists.html) and come\nback to this one.\n\n## Goals\n\nIn this tutorial we&#39;ll look into some last bits of Glam&#39;s API that were not\ncovered by the other tutorials. You&#39;ll learn how to deal with new lines and\nhow to flexibly display documents to get the most out of a line.\n\nAs a running example we&#39;re going to be pretty printing error messages.\nIt&#39;s always a joy to find a compiler with beautifully formatted error messages,\nwith contextual hints that can point to specific errors in your source code.\nIf you&#39;ve ever programmed with Elm or Rust, you&#39;ll know what I&#39;m talking about.\n\nIn this tutorial we&#39;re not concerned with _how_ the errors are generated - we&#39;re\nnot writing a compiler after all - we&#39;re just going to pretty print them.\nThe reference data structure I&#39;m going to use is this one:\n\n```gleam\ntype Span {\n  Span(line: Int, column_start: Int, column_end: Int)\n}\n\ntype Error {\n  Error(code: String, name: String, message: String, span: Span)\n}\n```\n\nIt&#39;s quite bare-bones but more than enough for our needs:\n\n- An `Error` has an error code, a name, a long form message, and a `Span` that\n  is used to tell which part of the source code it is referencing\n- A `Span` is used to point to a specific piece of the source code: it\n  references a line and can span multiple columns\n\nLet&#39;s look at an example error to make things clearer:\n\n```gleam\nlet source_code = &quot;fn greet(message: String) -&gt; Nil {\n  println(\\&quot;Hello\\&quot; &lt;&gt; message)\n}&quot;\nlet message = &quot;The name `println` is not in scope here.\nDid you mean to use `io.println`?&quot;\nlet unknown_variable = Error(&quot;E011&quot;, &quot;Unknown variable&quot;, message, Span(1, 2, 8))\n```\n\nHow should this error look? This is what we&#39;re going for look-wise:\n\n```text\n[ E011 ]: Unknown variable\n2 |   println(&quot;Hello&quot; &lt;&gt; message)\n      ┬─────\n      ╰─ The name `println` is not in scope here.\n         Did you mean to use `io.println`?\n```\n\nPretty nice, isn&#39;t it? There are a couple rules that the pretty printed error\nmessage should follow:\n\n- There should be a header with the error code and its name\n- The error&#39;s line should be displayed and the columns covered by the span\n  should be underlined\n- There should be a tooltip starting from the underline that displays the error\n  message\n- The error message&#39;s text should nicely wrap if the space is not enough to fit\n  on a single line\n\nTo get a sense of how the text wrapping should work, let&#39;s have a look at an\nerror message with a longer description:\n\n```gleam\nlet source_code = &quot;fn greet(message: String) -&gt; Nil {\n  println(\\&quot;Hello\\&quot; &lt;&gt; message)\n}&quot;\nlet message = &quot;This function is unused!\nYou can safely remove it or make it public with the `pub` keyword.&quot;\nlet unused_function = Error(&quot;E001&quot;, &quot;Unused function&quot;, message, Span(0, 3, 7))\n```\n\n```text\n[ E001 ]: Unused function\n1 | fn greet(message: String) -&gt; Nil {\n       ┬───\n       ╰─ This function is unused!\n          You can safely remove it or make it public with\n          the `pub` keyword.\n```\n\nAs you can see here, the pretty printer splits the error message&#39;s second line\nin order to fit the text in the allowed width (60 characters in this case).\nAlso, it was considerate enough to leave the new line that was explicitly used.\n\nIt may look like a simple example but it&#39;s actually remarkable: manually\nsplitting the error messages would be impossible since they could start at any\npoint of the source code, being indented by an arbitrary amount of spaces.\nOn the other hand, hand-writing a function to nicely split the text while also\ntaking into account the indentation is a pain (trust me,\n[I&#39;ve](https://github.com/giacomocavalieri/prequel/blob/ec293aa066ab2f5e968c994cdb6026058f561470/src/prequel/internals/report.gleam#L381-L444)\n[tried](https://github.com/giacomocavalieri/prequel/blob/ec293aa066ab2f5e968c994cdb6026058f561470/src/prequel/internals/extra/string.gleam#L23-L48)).\n\nHowever, by using a `Document`, the compiler author can forget about these\ndetails and simply focus on writing great error messages!\nEverything that has to do with tedious text manipulation will be taken care of\nautomatically by Glam.\n\n## Flexibly splitting a text\n\nLet&#39;s jump straight into the problem of splitting a text in a flexible way:\nwe want the words to wrap on a new line only if the text wouldn&#39;t fit on a\nsingle one.\n\nThis kind of _wrapping_ behaviour, however, cannot be achieved by simply using\nthe functions I showed you in the other tutorials. Let me show you why:\n\n### The problem with `doc.break`s\n\nIn all the other tutorials, when we needed to add a space that could be broken\nby the pretty printer, we generally used `doc.break` (or `doc.space` and\n`doc.soft_break`, which are implemented as a call to `doc.break`).\n\nYou may wonder why we cannot use the same approach here as well. I think an\nexample will be a perfect starting point to get why:\n\n```gleam\nlet text =\n  [&quot;Lorem&quot;, &quot;ipsum&quot;, &quot;dolor&quot;, &quot;sit&quot;, &quot;amet&quot;]\n  |&gt; list.map(doc.from_string)\n  |&gt; doc.join(with: doc.space)\n  |&gt; doc.group\n\ntext |&gt; doc.to_string(20)\n// -&gt;\n// Lorem\n// ipsum\n// dolor\n// sit\n// amet\n```\n\nAs you can see, every single space was broken into a new line. This is the\nexpected behaviour of a group: it acts a lot like a database transaction\nwhere either no `doc.break` gets broken or _every single one_ inside the group\ngets split on a new line.\n\n&gt; I have to thank Louis for the nice analogy with database transactions, it&#39;s a\n&gt; great way of presenting `doc.group`&#39;s behaviour with regard to breaks.\n&gt;\n&gt; When I first got into pretty printing and was struggling with the grouping\n&gt; behaviour, this way of putting it really made it click for me!\n\nI hope you can now see why this kind of _all or nothing_ behaviour cannot be\nused to get the nice word wrapping effect we are looking for.\n\n### Enters `doc.flex_break`\n\nTo solve this problem, Glam has another function called `doc.flex_break` which\nbehaves a lot like a `doc.break`: it is another possible breaking point for the\npretty printer, and it can either be displayed as its first argument (if kept on\na single line) or as its second one (if the pretty printer decides to split the\ndocument).\n\nThe fundamental difference lies in the way the pretty printer decides to break\na `doc.flex_break` inside a group: even if the pretty printer decided to\nsplit all normal `doc.break`s inside it, a `doc.flex_break` is never split\nunless absolutely necessary. This may sound confusing at first, so let&#39;s have a\nlook at an example:\n\n```gleam\nlet text =\n  [\n    doc.from_string(&quot;one&quot;),\n    doc.space,\n    doc.from_string(&quot;two&quot;),\n    doc.space,\n    doc.from_string(&quot;three&quot;),\n    doc.flex_break(&quot; &quot;, &quot;&quot;),\n    doc.from_string(&quot;four&quot;),\n  ]\n  |&gt; doc.join\n  |&gt; doc.group\n```\n\nLet&#39;s see what happens if we give the document enough space to fit on one line:\n\n```gleam\ndoc.to_string(text, 80)\n// -&gt; one two three four\n```\n\nSo far so good, we could have expected this result. Now let&#39;s see what happens\nif we reduce the line width:\n\n```gleam\ndoc.to_string(text, 10)\n// -&gt;\n// one\n// two\n// three four\n```\n\nThis may be surprising so let&#39;s work it out step-by-step:\n\n- The pretty printer runs into the outermost group and it tries to fit\n  it on a single line. However, the string `&quot;one two three four&quot;` is longer\n  than 10 characters\n- The pretty printer commits to breaking every single space inside the group,\n  so when it runs into the first two spaces those get split on new lines\n- Now the pretty printer has to display the last bit. As I&#39;ve described before,\n  all `doc.flex_break` never get split automatically like normal breaks.\n  Now the pretty printer can decide to keep the last `doc.flex_break` on a\n  single line if the remaining document fits on a single line\n- The string `&quot;three four&quot;` fits perfectly in 10 characters so it is not split.\n  If that were a `doc.break` (remember that a `doc.space` is just\n  `doc.break(&quot; &quot;, &quot;&quot;)`) the pretty printer would have been forced to split it\n\n### Splitting the text\n\nI hope that now you have an intuition on how we could use `doc.flex_break` to\ndisplay a text whose words can wrap on new lines as needed:\n\n```gleam\nfn flexible_text(text: String) -&gt; Document {\n  string.split(line, on: &quot; &quot;)\n  |&gt; list.map(doc.from_string)\n  |&gt; doc.join(with: doc.flex_space)\n  |&gt; doc.group\n}\n```\n\nWe replace every single whitespace in the string with a `doc.flex_space` (which\nis just a call to `doc.flex_break(&quot; &quot;, &quot;&quot;)`). Now if we try to display an\nexample text with different line widths we&#39;ll get a nice-looking result:\n\n```gleam\nlet message = &quot;I&#39;m a flexible text and I can wrap on multiple lines!&quot;\nflexible_text(message) |&gt; doc.to_string(30)\n// -&gt;\n// I&#39;m a flexible text and I can\n// wrap on multiple lines!\n\nflexible_text(message) |&gt; doc.to_string(15)\n// -&gt;\n// I&#39;m a flexible\n// text and I can\n// wrap on\n// multiple lines!\n```\n\n### Gracefully handling new lines\n\nThere&#39;s only one small problem with this solution: we&#39;re not actually handling\nnew lines the way one would expect. You can immediately notice this if you add\nsome nesting to the flexible text:\n\n```gleam\nflexible_test(&quot;this is a text\\nthat is split flexibly&quot;)\n// ^-- notice how there&#39;s a new line here\n|&gt; doc.nest(by: 2)\n|&gt; doc.to_string(10)\n// -&gt;\n// this is a\n//   text\n// that\n//   is split\n//   flexibly\n```\n\nThe pretty printer behaved in a kind of surprising way and didn&#39;t add\nany indentation before the new line... and rightfully so! The pretty printer is\ndoing the correct thing here.\n\n&gt; When we build a document using `doc.from_string` the pretty printer will\n&gt; _never_ split that string, even if it where to exceed the line limit. If it\n&gt; tried to be smart and split the raw strings passed to `doc.from_string`, it\n&gt; could end up doing more harm than good.\n&gt;\n&gt; Imagine you&#39;re writing a pretty printer for a programming language and the\n&gt; pretty printer randomly inserted new lines producing an invalid program. That\n&gt; would be quite annoying.\n&gt;\n&gt; Instead, we want to be really explicit about where the pretty printer is\n&gt; allowed to add breaks, hence why the `doc.break` and `doc.flex_break`\n&gt; functions.\n\nWhen we run into a `&quot;\\n&quot;` _we - human beings -_ expect to see a new line\nand some indentation, but the pretty printer just sees a\n`doc.from_string(&quot;\\n&quot;)` and doens&#39;t perform any kind of inspection of the\nstring&#39;s content.\n\nOk this is all nice and cool, but how can we actually add new lines that play\nnicely with the pretty printer and can also get indented automatically?\n\nThis is the job for `doc.line`: when the pretty printer runs into it, it simply\nrenders it as a new line. The only difference is that now the pretty printer\n_knows_ that is is going on a new line and it can also add indentation!\n\n### Handling new lines in the flexible text\n\nWe want to handle new lines gracefully in the flexible text we&#39;re building so\nthat, if we ever need to indent it ~~(foreshadowing)~~, it will still be nicely\nrendered by the pretty printer:\n\n```gleam\nfn flexible_text(text: String) -&gt; Document {\n  let to_flexible_line = fn(line) {\n    string.split(line, on: &quot; &quot;)\n    |&gt; list.map(doc.from_string)\n    |&gt; doc.join(with: doc.flex_space)\n    |&gt; doc.group\n  }\n\n  string.split(text, on: &quot;\\n&quot;)\n  |&gt; list.map(to_flexible_line)\n  |&gt; doc.join(with: doc.line)\n  // ^-- replace every `&quot;\\n&quot;` with a `doc.line`\n  |&gt; doc.group\n}\n```\n\n&gt; The bottom line is (sorry for the horrible pun, I had to do it): if we ever\n&gt; need to join documents together separating those with a new line, we should\n&gt; always use `doc.line` so that the pretty printer can handle it correctly and\n&gt; also add the needed indentation.\n\n## Displaying the other bits\n\nSplitting the text was the hardest part, now we can focus on other bits of the\nerror message that shouldn&#39;t be too complex.\n\n### The error&#39;s header\n\nThe error&#39;s header is quite straightforward and doesn&#39;t require any careful\nhandling.\nWe have to concatenate a couple of strings to get the desired look and turn\nthose into a fixed document with `doc.from_string`:\n\n```gleam\nfn header_doc(code: String, name: String) -&gt; Document {\n  [&quot;[&quot;, code, &quot;]:&quot;, name]\n  |&gt; string.join(with: &quot; &quot;)\n  |&gt; doc.from_string\n}\n```\n\n### The code line\n\nWe want to display the line where the error takes place without changing it.\nThe pretty printer should never try to split it; imagine if the code in your\nerror messages looked different from the actual code you&#39;ve written, that would\nbe rather confusing!\n\nThat also makes things easier for us, we just have to prefix the line with its\nnumber and we can call it a day:\n\n```gleam\nfn line_doc(source_code: String, line_number: Int) -&gt; Document {\n  let source_code_lines = string.split(source_code, on: &quot;\\n&quot;)\n  let assert Ok(line) = list.at(source_code_lines, line_number)\n  doc.from_string(line_prefix(line_number) &lt;&gt; line)\n}\n\nfn line_prefix(line_number: Int) -&gt; String {\n  int.to_string(line_number + 1) &lt;&gt; &quot; | &quot;\n  // ^-- pay attention to the off-by-one errors, lines are 0 based for the error\n  //     type but we want them to start from 1 for the user!\n}\n```\n\n&gt; Notice how we do a `let assert` when extracting the code line from the source\n&gt; code lines. This is an assumption that I made to make things a bit simpler: we\n&gt; expect that the compiler will always produce a valid error; that is, that any\n&gt; error that we&#39;re tasked with printing refers to an actual line of the source\n&gt; code.\n\n### The underlined pointer\n\nWhen displaying the pointer, we have to be careful with its size: it&#39;s going to\nbe composed of a single `&quot;┬&quot;` character and a sequence of dashes to underline\nthe whole span.\nWe can write a helper function that can create such pointer given the length it\nshould cover:\n\n```gleam\nfn underlined_pointer(length: Int) -&gt; Document {\n  [\n    doc.from_string(&quot;┬&quot; &lt;&gt; string.repeat(&quot;─&quot;, length - 1)),\n    doc.line,\n    doc.from_string(&quot;╰─ &quot;)\n  ]\n  |&gt; doc.concat\n}\n```\n\n&gt; Notice how we use a `doc.line` here as well. Remember: you should use a\n&gt; `doc.line` every time you need to join documents separated by a newline!\n\n## Putting everything together\n\nWe&#39;ve written a bunch of small self-contained functions for generating the bits\nand pieces needed by the final error message. Now it&#39;s time to put everything\ntogether! Our final function&#39;s type should turn an `Error` into a `Document`:\n\n```gleam\nfn error_to_doc(source_code: String, error: Error) -&gt; Document {\n  todo as &quot;let&#39;s put everything together!&quot;\n}\n```\n\nFirst we&#39;ll just `doc.concat` every single document and then start tweaking from there:\n\n```gleam\nfn error_to_doc(source_code: String, error: Error) -&gt; Document {\n  let underline_size = error.span.column_end - error.span.column_start + 1\n\n  [\n    header_doc(error.code, error.name),\n    doc.line,\n    line_doc(source_code, error.span.line),\n    doc.line,\n    underlined_pointer(underline_size),\n    doc.line,\n    flexible_text(error.message),\n  ]\n  |&gt; doc.concat\n}\n```\n\nLet&#39;s see how this works with an example error:\n\n```gleam\nerror_to_doc(source_code, unknown_variable)\n|&gt; doc.to_string(30)\n// -&gt;\n// [ E011 ]: Unknown variable\n// 2 |   println(&quot;Hello&quot; &lt;&gt; message)\n// ┬──────\n// ╰─ \n// The name `println` is not in\n// scope here.\n// Did you mean to use\n// `io.println`?\n```\n\nNot bad! It just needs some finishing touches and it should be good to go:\n\n- The underline should start right under the error location, right now it&#39;s\n  at the beginning of its line with no indentation\n- The text should start at the same line of the pointer and not on a new line\n- The text should be indented so that it is always to the right of the pointer\n\n### Joining the text and the pointer\n\nWe need to make sure the error message actually starts at the right of the\npointer and not on a new line. In order to do so we can `doc.concat` the two\ndocuments together.\n\nI don&#39;t want the body of the `error_to_doc` function to get too big and out of\nhand so I&#39;ll handle this in an helper function:\n\n```gleam\nfn message_doc(message: String, length: Span) -&gt; Document {\n  [\n    underlined_pointer(span_length),\n    flexible_text(message),\n  ]\n  |&gt; doc.concat\n}\n```\n\nLet&#39;s try this function out and see if it&#39;s missing something:\n\n```gleam\nmessage_doc(&quot;a long error message that is going to be wrapped&quot;, 3)\n|&gt; doc.to_string(30)\n// -&gt;\n// ┬──\n// ╰─ a long error message that\n// is going to be wrapped\n```\n\nAlmost there, the text needs to be indented whenever it gets wrapped. The right\namount of indentation is 3 spaces:\n\n```gleam\nfn message_doc(message: String, length: Span) -&gt; Document {\n  [\n    underlined_pointer(span_length),\n    flexible_text(message)\n    |&gt; doc.nest(by: 3),\n  ]\n  |&gt; doc.concat\n}\n\nmessage_doc(&quot;a long error message that is going to be wrapped&quot;, 3)\n|&gt; doc.to_string(30)\n// -&gt;\n// ┬──\n// ╰─ a long error message that\n//    is going to be wrapped\n```\n\nWe can now update `error_to_doc` to use this function:\n\n```gleam\nfn error_to_doc(source_code: String, error: Error) -&gt; Document {\n  let underline_size = error.span.column_end - error.span.column_start + 1\n\n  [\n    header_doc(error.code, error.name),\n    doc.line,\n    line_doc(source_code, error.span.line),\n    doc.line,\n    message_doc(error.message, underline_size),\n  ]\n  |&gt; doc.concat\n}\n```\n\n### Aligning the pointer\n\nNow we just need to move the document produced by `message_doc` to the right to\nalign it with the starting point of the error message. In order to do that we\ncan use `doc.nest` like usual.\n\nWhat&#39;s the right amount of indentation needed to align the pointer? If we didn&#39;t\nadd the line number to the left of the code line, we could simply use the span&#39;s\n`column_start` as the indentation.\n\nSince we added a prefix we also need to take its size into account when aligning\nthe error message to its starting point:\n\n```text\n1 | fn greet(message: String) -&gt; Nil {\n^^^^~~~ → this is the span&#39;s `column_start`\n   ⤷ this is the size of the prefix we added\n```\n\nWe can achieve this by making a little change to the `line_doc` function to\nmake it returns the size of the added prefix as well:\n\n```gleam\nfn line_doc(source_code: String, line_number: Int) -&gt; #(Document, Int) {\n  let source_code_lines = string.split(source_code, on: &quot;\\n&quot;)\n  let assert Ok(line) = list.at(source_code_lines, line_number)\n  let prefix = line_prefix(line_number)\n  let prefix_size = string.length(prefix)\n  #(doc.from_string(prefix &lt;&gt; line), prefix_size)\n  // ^-- now we also return the size of the prefix we&#39;ve added\n  //    to the left of the string\n}\n```\n\nWith this we can complete the `error_to_doc` function:\n\n```gleam\nfn error_to_doc(source_code: String, error: Error) -&gt; Document {\n  let underline_size = error.span.column_end - error.span.column_start + 1\n  let #(line_doc, prefix_size) = line_doc(source_code, error.span.line)\n  let message_indentation = error.span.column_start + prefix_size\n\n  [\n    header_doc(error.code, error.name),\n    doc.line,\n    line_doc,\n    [doc.line, message_doc(error.message, underline_size)]\n    // ^-- the line needs to be indented as well, so it&#39;s in the same\n    //     concat as the message_doc\n    |&gt; doc.concat\n    |&gt; doc.nest(by: message_indentation),\n  ]\n  |&gt; doc.concat\n}\n```\n\n## A bird&#39;s eye view\n\nBefore closing this tutorial, let&#39;s take one last step back to get an overall\nview of the code we wrote:\n\n```gleam\nfn error_to_doc(source_code: String, error: Error) -&gt; Document\nfn header_doc(code: String, name: String) -&gt; Document\nfn line_doc(source_code: String, line_number: Int) -&gt; #(Document, Int)\nfn message_doc(message: String, length: Int) -&gt; Document\nfn underlined_pointer(length: Int) -&gt; Document\n```\n\nWe ended up writing a bunch of small, self-contained functions and then managed\nto join those together with a little effort in the `error_to_doc` function.\n\nWhat&#39;s truly remarkable is how easy it was to build every single piece in\nisolation, without having to worry about the context where it was going to be\nused.\n\nTake as an example the `underlined_pointer`: we built it ignoring the fact that\nit was going to be indented and composed with a flexible text in the end.\nAt the same time, the `message_doc` didn&#39;t have to worry about the fact that we\nwould later need to align it with the error&#39;s starting point.\n\n## Recap\n\nWe covered some bits of Glam&#39;s API that were not considered in other tutorials:\n\n- `doc.line` as a way to split documents on new lines, playing nicely with the\n  pretty printer\n- `doc.flex_break` as a way to flexibly split groups without having to commit\n  to the all-or-nothing behaviour of normal `doc.break`s\n\nAt the same time you also managed to display a beautiful error message; that&#39;s\nan achievement not to be underestimated!\n","url":"pretty_printing_lovely_error_messages.html"},{"doc":"glam/doc","title":"Document","content":"pub opaque type Document\n A document that can be pretty printed with `to_string`.\n \n\nSynonyms:\nglam/doc.Document\nglam/doc Document","url":"glam/doc.html#Document"},{"doc":"glam/doc","title":"empty","content":"pub const empty: Document = Concat([])\n An empty document that is printed as an empty string.\n \n ## Examples\n \n \n empty |&gt; to_string(80)\n // -&gt; &quot;&quot;\n \n \nSynonyms:\nglam/doc.empty\nglam/doc empty","url":"glam/doc.html#empty"},{"doc":"glam/doc","title":"flex_space","content":"pub const flex_space: Document = FlexBreak(&quot; &quot;, &quot;&quot;)\n A document that is rendered as a single whitespace `&quot; &quot;` but can be broken\n by the pretty printer into newlines instead.\n Instead of splitting all spaces or no spaces (like it would do with\n `space` inside a group), the pretty printer can decide on a space-by-space\n basis.\n \n If a `flex_space` inside a group needs to be broken to fit it on multiple\n lines, following `flex_spaces` may end up not being broken.\n \n This is a shorthand for `flex_break(&quot; &quot;, &quot;&quot;)`.\n \n ## Examples\n \n \n let message =\n   [&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]\n   |&gt; list.map(from_string)\n   |&gt; join(with: flex_space)\n   |&gt; group\n \n to_string(message, 80)\n // -&gt; &quot;Gleam is fun!&quot;\n \n to_string(message, 10)\n // -&gt; &quot;Gleam is\n fun!&quot;\n \n \n to_string(message, 4)\n // -&gt; &quot;Gleam\n // is\n // fun!&quot;\n \nSynonyms:\nglam/doc.flex_space\nglam/doc flex_space","url":"glam/doc.html#flex_space"},{"doc":"glam/doc","title":"line","content":"pub const line: Document = Line(1)\n A document that is always printed as a single new line.\n \n ## Examples\n \n \n line |&gt; to_string(80)\n // -&gt; &quot;\\n&quot;\n \n \nSynonyms:\nglam/doc.line\nglam/doc line","url":"glam/doc.html#line"},{"doc":"glam/doc","title":"soft_break","content":"pub const soft_break: Document = Break(&quot;&quot;, &quot;&quot;)\n A document that is always rendered as an empty string but can act as a\n breaking point for the pretty printer.\n \n This is a shorthand for `break(&quot;&quot;, &quot;&quot;)`.\n \n ## Examples\n \n \n let doc = [from_string(&quot;soft&quot;), soft_break, from_string(&quot;break&quot;)]\n \n doc |&gt; to_string(80)\n // -&gt; &quot;softbreak&quot;\n \n doc |&gt; to_string(5)\n // -&gt; &quot;soft\n // break&quot;\n \n \nSynonyms:\nglam/doc.soft_break\nglam/doc soft_break","url":"glam/doc.html#soft_break"},{"doc":"glam/doc","title":"space","content":"pub const space: Document = Break(&quot; &quot;, &quot;&quot;)\n A document that is rendered as a single whitespace `&quot; &quot;` but can be broken\n by the pretty printer into newlines instead.\n \n This is a shorthand for `break(&quot; &quot;, &quot;&quot;)`.\n \n ## Examples\n \n \n let doc =\n   [&quot;pretty&quot;, &quot;printed&quot;]\n   |&gt; list.map(from_string)\n   |&gt; join(with: space)\n \n doc |&gt; to_string(80)\n // -&gt; &quot;pretty printed&quot;\n \n doc |&gt; to_string(10)\n // -&gt; &quot;pretty\n // printed&quot;\n \n \nSynonyms:\nglam/doc.space\nglam/doc space","url":"glam/doc.html#space"},{"doc":"glam/doc","title":"append","content":"pub fn append(\n  to first: Document,\n  doc second: Document,\n) -&gt; Document\n Joins a document into the end of another.\n \n ## Examples\n \n \n from_string(&quot;pretty&quot;)\n |&gt; append(from_string(&quot; printer&quot;))\n |&gt; to_string(80)\n // -&gt; &quot;pretty printer&quot;\n \n \nSynonyms:\nglam/doc.append\nglam/doc append","url":"glam/doc.html#append"},{"doc":"glam/doc","title":"append_docs","content":"pub fn append_docs(\n  first: Document,\n  docs: List(Document),\n) -&gt; Document\n Joins multiple documents into the end of another.\n \n This is a shorthand for `append(to: first, doc: concat(docs))`.\n \n ## Examples\n \n \n from_string(&quot;pretty&quot;)\n |&gt; append_docs([\n   from_string(&quot;printing&quot;),\n   space,\n   from_string(&quot;rocks!&quot;),\n ])\n |&gt; to_string(80) \n // -&gt; &quot;pretty printing rocks!&quot;\n \n \nSynonyms:\nglam/doc.append_docs\nglam/doc append_docs","url":"glam/doc.html#append_docs"},{"doc":"glam/doc","title":"break","content":"pub fn break(unbroken: String, broken: String) -&gt; Document\n A document after which the pretty printer can insert a new line.\n A newline is added after a `break` document if the `group` it&#39;s part of\n could not be rendered on a single line.\n \n If the pretty printer decides to add a newline after `break` it will be\n rendered as its second argument, otherwise as its first argument.\n \n ## Examples\n \n \n let message =\n   [from_string(&quot;pretty&quot;), break(&quot;•&quot;, &quot;↩&quot;), from_string(&quot;printed&quot;)]\n   |&gt; concat\n   |&gt; group\n \n message |&gt; to_string(20)\n // -&gt; &quot;pretty•printed&quot;\n \n message |&gt; to_string(10)\n // -&gt; &quot;pretty↩\n // printed&quot;\n \n \nSynonyms:\nglam/doc.break\nglam/doc break","url":"glam/doc.html#break"},{"doc":"glam/doc","title":"concat","content":"pub fn concat(docs: List(Document)) -&gt; Document\n Joins a list of documents into a single document.\n \n The resulting pretty printed document would be the same as pretty printing\n each document separately and concatenating it together with `&lt;&gt;`:\n \n \n docs |&gt; concat |&gt; to_string(n) ==\n docs |&gt; list.map(to_string(n)) |&gt; string.concat\n \n \n ## Examples\n \n \n [&quot;pretty&quot;, &quot; &quot;, &quot;printed&quot;]\n |&gt; list.map(from_string)\n |&gt; concat\n |&gt; to_string(80)\n // -&gt; &quot;pretty printed&quot;\n \n \nSynonyms:\nglam/doc.concat\nglam/doc concat","url":"glam/doc.html#concat"},{"doc":"glam/doc","title":"concat_join","content":"pub fn concat_join(\n  docs: List(Document),\n  with separators: List(Document),\n) -&gt; Document\n Joins a list of documents into a single one by inserting the given\n separators between each existing document.\n \n This is a shorthand for `join(docs, concat(separators))`.\n \n ## Examples\n \n \n [&quot;wow&quot;, &quot;so&quot;, &quot;many&quot;, &quot;commas&quot;]\n |&gt; list.map(from_string)\n |&gt; concat_join([from_string(&quot;,&quot;), space])\n |&gt; to_string(80)\n // -&gt; &quot;wow, so, many, commas&quot;\n \n \nSynonyms:\nglam/doc.concat_join\nglam/doc concat_join","url":"glam/doc.html#concat_join"},{"doc":"glam/doc","title":"flex_break","content":"pub fn flex_break(unbroken: String, broken: String) -&gt; Document\n A document after which the pretty printer can insert a new line.\n The difference with a simple `break` is that, the pretty printer will decide\n wether to add a new line or not on a space-by-space basis.\n \n While all the `break` inside a group are broken or not, some `flex_breaks`\n may be broken and some not, depending wether the document can fit on a\n single line or not. Hence the name _flex_.\n \n If the pretty printer decides to add a newline after `flex_break` it will be\n rendered as its first argument, otherwise as its first argument.\n \n ## Examples\n \n \n let message =\n   [from_string(&quot;pretty&quot;), from_string(&quot;printed&quot;), from_string(&quot;string&quot;)]\n   |&gt; join(with: flex_break(&quot;•&quot;, &quot;↩&quot;))\n   |&gt; group\n \n message |&gt; to_string(80)\n // -&gt; &quot;pretty•printed•string&quot;\n \n message |&gt; to_string(20)\n // -&gt; &quot;pretty•printed↩\n // string&quot;\n \n \nSynonyms:\nglam/doc.flex_break\nglam/doc flex_break","url":"glam/doc.html#flex_break"},{"doc":"glam/doc","title":"force_break","content":"pub fn force_break(doc: Document) -&gt; Document\n Forces the pretty printer to break all the `break`s of the outermost\n document. This still _has no effect on `group`s_ as the pretty printer will\n always try to put them on a single line before splitting them.\n \n ## Examples\n \n \n [from_string(&quot;pretty&quot;), break(&quot;•&quot;, &quot;↩&quot;), from_string(&quot;printed&quot;)]\n |&gt; concat\n |&gt; force_break\n |&gt; group\n |&gt; to_string(100)\n // -&gt; &quot;pretty↩\n // printed&quot;\n \n \nSynonyms:\nglam/doc.force_break\nglam/doc force_break","url":"glam/doc.html#force_break"},{"doc":"glam/doc","title":"from_string","content":"pub fn from_string(string: String) -&gt; Document\n Turns a string into a document.\n \n ## Examples\n \n \n &quot;doc&quot; |&gt; from_string |&gt; to_string(80)\n // -&gt; &quot;doc&quot;\n \n \nSynonyms:\nglam/doc.from_string\nglam/doc from_string","url":"glam/doc.html#from_string"},{"doc":"glam/doc","title":"group","content":"pub fn group(doc: Document) -&gt; Document\n Allows the pretty printer to break the `break` documents inside the given\n group.\n \n When the pretty printer runs into a group it first tries to render it on a\n single line, displaying all the breaks as their first argument.\n If the group fits this is the final pretty printed result.\n \n However, if the group does not fit on a single line _all_ the `break`s\n inside that group are rendered as their second argument and immediately\n followed by a newline.\n \n Any nested group is considered on its own and may or may not be split,\n depending if it fits on a single line or not. So, even if the outermost\n group is broken, its nested groups may still end up on a single line.\n \n ## Examples\n \n \n let food =\n   [&quot;lasagna&quot;, &quot;ravioli&quot;, &quot;pizza&quot;]\n   |&gt; list.map(from_string) |&gt; join(with: space) |&gt; group\n let message = \n   [from_string(&quot;Food I love:&quot;), space, food] |&gt; concat |&gt; group\n \n message |&gt; to_string(80)\n // -&gt; &quot;Food I love: lasagna ravioli pizza&quot;\n \n message |&gt; to_string(30)\n // -&gt; &quot;Food I love:\n // lasagna ravioli pizza&quot;\n // ^-- After splitting the outer group, the inner one can fit\n //     on a single line so the pretty printer does not split it\n \n message |&gt; to_string(20)\n // &quot;Food I love:\n // lasagna\n // ravioli\n // pizza&quot;\n // ^-- Even after splitting the outer group, the inner one wouldn&#39;t\n //     fit on a single line, so the pretty printer splits that as well\n \n \nSynonyms:\nglam/doc.group\nglam/doc group","url":"glam/doc.html#group"},{"doc":"glam/doc","title":"join","content":"pub fn join(\n  docs: List(Document),\n  with separator: Document,\n) -&gt; Document\n Joins a list of documents inserting the given separator between\n each existing document.\n \n ## Examples\n \n \n let message =\n   [&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]\n   |&gt; list.map(from_string)\n   |&gt; join(with: space)\n \n message |&gt; to_string(80)\n // -&gt; &quot;Gleam is fun!&quot;\n \n \nSynonyms:\nglam/doc.join\nglam/doc join","url":"glam/doc.html#join"},{"doc":"glam/doc","title":"lines","content":"pub fn lines(size: Int) -&gt; Document\n A document that is always printed as a series of consecutive newlines.\n \n ## Examples\n \n \n lines(3) |&gt; to_string(80)\n // -&gt; &quot;\\n\\n\\n&quot;\n \n \nSynonyms:\nglam/doc.lines\nglam/doc lines","url":"glam/doc.html#lines"},{"doc":"glam/doc","title":"nest","content":"pub fn nest(doc: Document, by indentation: Int) -&gt; Document\n Increases the nesting level of a document by the given amount.\n \n When the pretty printer breaks a group by inserting a newline, it also adds\n a whitespace padding equal to its nesting level.\n \n ## Examples\n \n \n let one = [space, from_string(&quot;one&quot;)] |&gt; concat |&gt; nest(by: 1)\n let two = [space, from_string(&quot;two&quot;)] |&gt; concat |&gt; nest(by: 2)\n let three = [space, from_string(&quot;three&quot;)] |&gt; concat |&gt; nest(by: 3)\n let list = [from_string(&quot;list:&quot;), one, two, three] |&gt; concat |&gt; group\n \n list |&gt; to_string(10)\n // -&gt; &quot;list:\n //  one\n //   two\n //    three&quot;\n \n\nSynonyms:\nglam/doc.nest\nglam/doc nest","url":"glam/doc.html#nest"},{"doc":"glam/doc","title":"nest_docs","content":"pub fn nest_docs(\n  docs: List(Document),\n  by indentation: Int,\n) -&gt; Document\n Joins together a list of documents and increases their nesting level.\n \n This is a shorthand for `nest(concat(docs), by: indentation)`.\n \n ## Examples\n \n \n [from_string(&quot;one&quot;), space, from_string(&quot;two&quot;)]\n |&gt; nest_docs(by: 2)\n |&gt; append(space)\n |&gt; append(from_string(&quot;three&quot;))\n |&gt; group\n |&gt; to_string(5)\n // -&gt;\n // one\n //   two\n // three\n \n \nSynonyms:\nglam/doc.nest_docs\nglam/doc nest_docs","url":"glam/doc.html#nest_docs"},{"doc":"glam/doc","title":"prepend","content":"pub fn prepend(\n  to first: Document,\n  doc second: Document,\n) -&gt; Document\n Prefixes a document to another one.\n \n ## Examples\n \n \n from_string(&quot;printed!&quot;)\n |&gt; prepend(from_string(&quot;pretty &quot;))\n |&gt; to_string(80)\n // -&gt; &quot;pretty printed!&quot;\n \n \nSynonyms:\nglam/doc.prepend\nglam/doc prepend","url":"glam/doc.html#prepend"},{"doc":"glam/doc","title":"prepend_docs","content":"pub fn prepend_docs(\n  first: Document,\n  docs: List(Document),\n) -&gt; Document\n Prefixes multiple documents to another one.\n \n This is a shorthand for `prepend(to: first, doc: concat(docs))`.\n \n ## Examples\n \n \n from_string(&quot;fun!&quot;)\n |&gt; prepend_docs([from_string(&quot;Gleam &quot;), from_string(&quot;is &quot;)])\n |&gt; to_string(80)\n // -&gt; &quot;Gleam is fun!&quot;\n \n \nSynonyms:\nglam/doc.prepend_docs\nglam/doc prepend_docs","url":"glam/doc.html#prepend_docs"},{"doc":"glam/doc","title":"to_string","content":"pub fn to_string(doc: Document, width: Int) -&gt; String\n Turns a document into a pretty printed string.\n The pretty printed process can be thought of as follows:\n - the pretty printer first tries to print every group on a single line\n - all the `break` documents are rendered as their first argument\n - if the string fits on the specified width this is the result\n - if the string does not fit on a single line the outermost group is split:\n   - all of its `break` documents are rendered as their second argument\n   - a newline is inserted after every `break`\n   - a padding of the given nesting level is added after every inserted\n     newline\n   - all inner groups are then considered on their own: the splitting of the\n     outermost group does not imply that the inner groups will be split as\n     well\n \n ## Examples\n \n For some examples of how pretty printing works for each kind of document you\n can have a look at the package documentation.\n There&#39;s also a\n [step-by-step tutorial](https://hexdocs.pm/glam/01_gleam_lists.html)\n that will guide you through the implementation of a simple pretty printer,\n covering most of the Glam API.\n \nSynonyms:\nglam/doc.to_string\nglam/doc to_string","url":"glam/doc.html#to_string"},{"doc":"glam/doc","title":"to_string_builder","content":"pub fn to_string_builder(\n  doc: Document,\n  width: Int,\n) -&gt; StringBuilder\n This function works like `to_string` but returns a `StringBuilder` instead\n of a `String`.\n \nSynonyms:\nglam/doc.to_string_builder\nglam/doc to_string_builder","url":"glam/doc.html#to_string_builder"},{"doc":"glam/doc","title":"glam/doc","content":"","url":"glam/doc.html"}]);